@@ -1,5 +1,18 @@
+//! Toipe is a terminal-based typing test application.
+//!
+//! Please see the [README](https://github.com/Samyak2/toipe/) for
+//! installation and usage instructions.
+//!
+//! Toipe provides an API to invoke it from another application or
+//! library. This documentation describes the API and algorithms used
+//! internally.
+//!
+//! See [`RawWordSelector`] if you're looking for the word selection
+//! algorithm.
+
 pub mod config;
 pub mod textgen;
+pub mod wordlists;
 
 use std::io::{stdout, StdinLock, Stdout, Write};
 use std::path::PathBuf;
@@ -14,20 +27,29 @@ use termion::{
     raw::{IntoRawMode, RawTerminal},
     style, terminal_size,
 };
-use textgen::WordSelector;
+use textgen::{RawWordSelector, WordSelector};
+use wordlists::{get_word_list, OS_WORDLIST_PATH};
 
+/// Typing test terminal UI and logic.
 pub struct Toipe {
     stdout: RawTerminal<Stdout>,
     text: String,
     words: Vec<String>,
-    word_selector: WordSelector,
+    word_selector: Box<dyn WordSelector>,
 }
 
+/// Represents any error caught in Toipe.
 #[derive(Debug)]
 pub struct ToipeError {
     pub msg: String,
 }
 
+/// Converts [`std::io::Error`] to [`ToipeError`].
+///
+/// This keeps only the error message.
+///
+/// TODO: there must be a better way to keep information from the
+/// original error.
 impl From<std::io::Error> for ToipeError {
     fn from(error: std::io::Error) -> Self {
         ToipeError {
@@ -37,10 +59,23 @@ impl From<std::io::Error> for ToipeError {
 }
 
 impl<'a> Toipe {
+    /// Initializes a new typing test on the standard output.
+    ///
+    /// See [`ToipeConfig`] for configuration options.
+    ///
+    /// Puts `stdout` in raw mode and initializes the word selector.
+    /// Also invokes [`Toipe::restart()`].
     pub fn new(config: ToipeConfig) -> Result<Self, ToipeError> {
         let stdout = stdout().into_raw_mode().unwrap();
 
-        let word_selector = WordSelector::new(PathBuf::from(config.wordlist_path))?;
+        let word_selector: Box<dyn WordSelector> =
+            if let Some(word_list) = get_word_list(config.wordlist.as_str()) {
+                Box::new(RawWordSelector::from_string(word_list.to_string())?)
+            } else if config.wordlist == "os" {
+                Box::new(RawWordSelector::from_path(PathBuf::from(OS_WORDLIST_PATH))?)
+            } else {
+                Box::new(RawWordSelector::from_path(PathBuf::from(config.wordlist))?)
+            };
 
         let mut toipe = Toipe {
             stdout,
@@ -54,6 +89,10 @@ impl<'a> Toipe {
         Ok(toipe)
     }
 
+    /// Make the terminal ready for the next typing test.
+    ///
+    /// Clears the screen, generates new words and displays them on the
+    /// UI.
     pub fn restart(&mut self) -> Result<(), ToipeError> {
         self.reset_screen()?;
 
@@ -95,15 +134,23 @@ impl<'a> Toipe {
         Ok(())
     }
 
-    pub fn flush(&mut self) -> Result<(), ToipeError> {
+    fn flush(&mut self) -> Result<(), ToipeError> {
         self.stdout.flush()?;
         Ok(())
     }
 
+    /// Start typing test by monitoring input keys.
+    ///
+    /// Must only be invoked after [`Toipe::restart()`].
+    ///
+    /// If the test completes successfully, returns a boolean indicating
+    /// whether the user wants to do another test and the
+    /// [`ToipeResults`] for this test.
     pub fn test(&mut self, stdin: StdinLock<'a>) -> Result<(bool, ToipeResults), ToipeError> {
         let mut input = Vec::<char>::new();
         let text: Vec<char> = self.text.chars().collect();
         let mut num_errors = 0;
+        let mut num_chars_typed = 0;
 
         let mut process_key = |key: Key| -> Result<bool, ToipeError> {
             match key {
@@ -120,6 +167,8 @@ impl<'a> Toipe {
                         return Ok(false);
                     }
 
+                    num_chars_typed += 1;
+
                     if text[input.len() - 1] == c {
                         write!(
                             self.stdout,
@@ -189,7 +238,8 @@ impl<'a> Toipe {
 
         let results = ToipeResults {
             num_words: self.words.len(),
-            num_chars: input.len(),
+            num_chars_typed,
+            num_chars_text: input.len(),
             num_errors,
             started_at,
             ended_at,
@@ -222,7 +272,7 @@ impl<'a> Toipe {
 
         let line = format!(
             "Mistakes: {} out of {} characters",
-            results.num_errors, results.num_chars
+            results.num_errors, results.num_chars_text
         );
         write!(
             self.stdout,
@@ -264,6 +314,31 @@ impl<'a> Toipe {
             line,
         )?;
 
+        let line = format!(
+            "Press {}r{} to restart, {}q{} to quit.",
+            color::Fg(color::Blue),
+            color::Fg(color::Reset),
+            color::Fg(color::Blue),
+            color::Fg(color::Reset)
+        );
+        // do not consider length of formatting characters
+        let zerowidths = format!(
+            "{}{}{}{}",
+            color::Fg(color::Blue),
+            color::Fg(color::Reset),
+            color::Fg(color::Blue),
+            color::Fg(color::Reset)
+        );
+        write!(
+            self.stdout,
+            "{}{}{}{}{}",
+            cursor::Goto(sizex / 2, sizey - 2),
+            cursor::Left((line.len() - zerowidths.len()) as u16 / 2),
+            style::Faint,
+            line,
+            style::NoFaint,
+        )?;
+
         // no cursor on results page
         write!(self.stdout, "{}", cursor::Hide)?;
         self.flush()?;
@@ -283,6 +358,11 @@ impl<'a> Toipe {
 }
 
 impl Drop for Toipe {
+    /// Resets terminal.
+    ///
+    /// Clears screen and sets the cursor to a non-blinking block.
+    ///
+    /// TODO: reset cursor to whatever it was before Toipe was started.
     fn drop(&mut self) {
         write!(
             self.stdout,
@@ -296,29 +376,50 @@ impl Drop for Toipe {
     }
 }
 
+/// Stores stats from a typing test.
 #[derive(Clone)]
 pub struct ToipeResults {
     num_words: usize,
-    num_chars: usize,
+    num_chars_typed: usize,
+    num_chars_text: usize,
     num_errors: usize,
     started_at: Instant,
     ended_at: Instant,
 }
 
 impl ToipeResults {
-    fn duration(&self) -> Duration {
+    /// Number of correctly typed letters
+    pub fn num_correct_chars(&self) -> usize {
+        self.num_chars_typed - self.num_errors
+    }
+
+    /// Duration of the test.
+    ///
+    /// i.e., the time between the user pressing the first key and them
+    /// typing the last letter.
+    pub fn duration(&self) -> Duration {
         self.ended_at.duration_since(self.started_at)
     }
 
+    /// Percentage of letters that were typed correctly.
     pub fn accuracy(&self) -> f64 {
-        1.0 - (self.num_errors as f64 / self.num_chars as f64)
+        self.num_correct_chars() as f64 / self.num_chars_typed as f64
     }
 
+    /// Speed in (correctly typed) characters per minute.
     pub fn cpm(&self) -> f64 {
-        self.num_chars as f64 / (self.duration().as_secs_f64() / 60.0)
+        self.num_correct_chars() as f64 / (self.duration().as_secs_f64() / 60.0)
     }
 
+    /// Speed in (correctly typed) words per minute.
+    ///
+    /// Measured as `cpm / (chars per word)` where `chars per word` is
+    /// measured as `(number of chars) / (number of words)`.
+    ///
+    /// Note: this is only an approximation because "correctly typed
+    /// words" is ambiguous when there could be a mistake in only one or
+    /// two letters of a word.
     pub fn wpm(&self) -> f64 {
-        self.num_words as f64 / (self.duration().as_secs_f64() / 60.0)
+        self.cpm() / (self.num_chars_text as f64 / self.num_words as f64)
     }
 }
